<?php
//
//	Module name      :	asyncjob.php
//	Version          :	1.0
//	Started on       :	Maj 6, 2005
//	Completed on     :	Maj 17, 2005
//
//	Module type      :	Library
//	Description      :	Functions to handle asynchronous job tables
//
//	Developed under  :	Php 4.1.2
//	Sw requirements  :	None
//
//	Parameters       :	None
//	Notes - usage    :	The Dsn used for handling jobs is the first parameter of all functions, must be
//						already initialized and must have the following tables:
//							CREATE TABLE Configuration (
//								Section varchar(32) NOT NULL default '',
//								Variable varchar(32) NOT NULL default '',
//								Value varchar(255) NOT NULL default '',
//								PRIMARY KEY  (Section,Variable)
//							);
//							CREATE TABLE Job (
//								Id int(10) unsigned NOT NULL auto_increment,
//								Description varchar(64) NOT NULL default '',
//								ProgPath varchar(255) NOT NULL default '',
//								Ts int(11) unsigned NOT NULL default '0',
//								UserId int(10) unsigned NOT NULL default '0',
//								PRIMARY KEY  (Id)
//							);
//							CREATE TABLE Periodic (
//								Id int(10) unsigned NOT NULL auto_increment,
//								LastQueuedTs int(11) unsigned NOT NULL default '0',
//								Period int(11) unsigned NOT NULL default '0',
//								Disabled tinyint(1) unsigned NOT NULL default '0',
//								Hidden tinyint(1) unsigned NOT NULL default '0',
//								IdJob int(10) unsigned NOT NULL default '0',
//								Param varchar(255) NOT NULL default '',
//								Ts int(11) unsigned NOT NULL default '0',
//								UserId int(10) unsigned NOT NULL default '0',
//								PRIMARY KEY  (Id)
//							);
//							CREATE TABLE Queue (
//								Id int(10) unsigned NOT NULL auto_increment,
//								Hidden tinyint(1) unsigned NOT NULL default '0',
//								IdJob int(10) unsigned NOT NULL default '0',
//								Param varchar(255) NOT NULL default '',
//								Ts int(11) unsigned NOT NULL default '0',
//								UserId int(10) unsigned NOT NULL default '0',
//								Status tinyint(4) unsigned NOT NULL default '0',
//								Result smallint(6) NOT NULL default '0',
//								Output text NOT NULL,
//								TsStart int(11) unsigned NOT NULL default '0',
//								Elapsed float(10,6) unsigned NOT NULL default '0.000000',
//								PRIMARY KEY  (Id)
//							);
//						To change table names, see $AJ_cfg_table_name, $AJ_job_table_name, $AJ_per_table_name,
//						$AJ_queue_table_name. To change section names in Configuration table see
//						$AJ_meta_sec_name.
//						Minimal contents of Configuration table is:
//							INSERT INTO Configuration (Section, Variable, Value) VALUES
//								('Locks', 'AJQueuedRun', '0'),
//								('MetaParam', 'JobId', '[JobId]'),
//								('MetaParam', 'LastQueuedTs', '[LastQueuedTs]'),
//								('MetaParam', 'PerId', '[PerId]'),
//								('MetaParam', 'Period', '[Period]'),
//								('MetaParam', 'QueueId', '[QueueId]');
//						For an example of Job descriptor and Job data, see $AJ_job_desc and $AJ_job_data.
//						The function AJQueuedRun can log data of ended jobs using DbLog module if it has been
//						activated (see $AJ_db_log); data are logged in a table named "AsyncJob" (see
//						$AJ_log_table_name to change its name), with definition:
//							CREATE TABLE AsyncJob (
//								Ts int(11) unsigned NOT NULL default '0',
//								Hidden tinyint(1) unsigned NOT NULL default '0',
//								ProgPath varchar(255) NOT NULL default '',
//								Param varchar(255) NOT NULL default '',
//								UserId int(10) unsigned NOT NULL default '0',
//								Result smallint(6) NOT NULL default '0',
//								Output text NOT NULL,
//								TsStart int(11) unsigned NOT NULL default '0',
//								Elapsed float(10,6) unsigned NOT NULL default '0.000000',
//								KEY Ts (Ts)
//							);
//						Other indexes can be added.
//


//--- Include files --------------------------------------------------------------------------------



require_once("odbccfg.php");
require_once("odbccfgutil.php");
require_once("dateutil.php");



//--- Public constants -----------------------------------------------------------------------------



define("AJ_MODULE_NAME", "asyncjob.php 1.0; on Maj 17, 2005 by J.Hubb.");

// Job status codes
define("AJ_ENDED",		0);
define("AJ_RUNNING",	1);
define("AJ_SCHEDULED",	2);
define("AJ_READY",		3);

// Timeout default values for a job to stay in queue after ending
define("AJ_OK_TIMEOUT",		120);
define("AJ_ERR_TIMEOUT",	3600);
define("AJ_NO_TIMEOUT",		-1);

// Error codes
define("AJ_NO_ERROR",			0);
define("AJ_DB_ERROR",			1);
define("AJ_JOB_ID_ERROR",		2);
define("AJ_PER_ID_ERROR",		3);
define("AJ_QUEUE_ID_ERROR",		4);
define("AJ_ALREADY_RUNNING",	5);
define("AJ_LAST_ERROR",			5);



//--- Public variables -----------------------------------------------------------------------------



// Table names used by AsyncJob
$AJ_cfg_table_name		= "Configuration";
$AJ_job_table_name		= "Job";
$AJ_per_table_name		= "Periodic";
$AJ_queue_table_name	= "Queue";

// Section names in Configuration table used by AsyncJob
$AJ_meta_sec_name		= "MetaParam";

// Example for job descriptor (boolean data are converted to integers; ReadOnly values are ignored if passed)
$AJ_job_desc = array(
		"Description"	=> "Job",		// ReadWrite; description of the job; not used
		"ProgPath"		=> "/bin/echo",	// ReadWrite; path of the program
		"Ts"			=> 0,			// ReadOnly; timestamp of descriptor insertion; set but not used
		"UserId"		=> 0			// ReadWrite; numeric Id of the User inserting this job; not used
		);

// Example for job data (boolean data are converted to integers; ReadOnly values are ignored if passed)
$AJ_job_data = array(
	// ------ Following indexes are used by all functions
		"Hidden"		=> 0,		// ReadWrite; not used
		"IdJob"			=> 1,		// ReadOnly; index in Job table
		"Param"			=> "",		// ReadWrite; job parameters
		"Ts"			=> 0,		// ReadOnly; timestamp of data insertion; set but not used
		"UserId"		=> 0,		// ReadWrite; numeric Id of the User inserting/enqueuing the job; not used
	// --- Following indexes are used only by functions handling periodic jobs
		"LastQueuedTs"	=> "0",		// ReadOnly; timestamp of last enqueuing of this job
		"Period"		=> "60",	// ReadWrite; minimum period to wait for rescheduling
		"Disabled"		=> 0,		// ReadWrite; when TRUE the periodic task won't be scheduled
	//--- Following indexes are used only by AJQueuedRead and AJQueuedReadAll functions
		"Status"		=> AJ_ENDED,// ReadOnly; execution status of the enqueued job
		"Result"		=> 0,		// ReadOnly; error code returned; not 0 means that job ended with errors
		"Output"		=> "",		// ReadOnly; output strings generated by the job
		"TsStart"		=> 0,		// ReadOnly; start timestamp of the job
		"Elapsed"		=> 0.0		// ReadOnly; duration of job running
		);

$AJ_db_log 			= FALSE;		// Change to TRUE to Log queries by DBLog module
$AJ_log_table_name = "AsyncJob";	// Name of the table used by DBLog



//--- Private constants ----------------------------------------------------------------------------



//--- Private variables ----------------------------------------------------------------------------



$_AJ_def_desc = array(
		"Description"	=> "EmptyJob",
		"ProgPath"		=> "/bin/false",	// Avoid O.S. problems
		"UserId"		=> 0
		);

$_AJ_def_data = array(
		"Period"	=> "60",
		"Disabled"	=> 0,
		"Hidden"	=> 0,
		"IdJob"		=> -1,			// Avoid random job reference
		"Param"		=> "",
		"UserId"	=> 0
		);

$_AJ_error_num = AJ_NO_ERROR;
$_AJ_error_msg = array(
	AJ_NO_ERROR			=> "No error",
	AJ_DB_ERROR			=> "Error raised from ODBC layer",
	AJ_JOB_ID_ERROR		=> "Job id not valid",
	AJ_PER_ID_ERROR		=> "Periodic job id not valid",
	AJ_QUEUE_ID_ERROR	=> "Queue job id not valid",
	AJ_ALREADY_RUNNING	=> "Procedure already in execution"
	);



//--- Public functions code ------------------------------------------------------------------------



function AJDescInsert(&$db_desc, $data)
// Inserts a new job descriptor, specifying its $data; returns the job id on success, FALSE on errors.
{
	global $_AJ_def_desc, $AJ_job_table_name, $_AJ_error_num;

	$data = array_merge($_AJ_def_desc, $data);			// Overwrite default values with new ones
	ODBCQuery($db_desc, "INSERT INTO $AJ_job_table_name " .
			"(Description, ProgPath, Ts, UserId) VALUES (" .
			"'" . addslashes($data["Description"]) . "', " .
			"'" . addslashes($data["ProgPath"])  . "', " .
			time() . ", " .
			$data["UserId"] . ")"
			);
	if ((ODBCResult() != ODBC_NO_ERROR) || (ODBCAffectedRows($db_desc) == 0))
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	list ($id) = ODBCSelectOneRow($db_desc, "SELECT LAST_INSERT_ID()");
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $id;
}



function AJDescModify(&$db_desc, $job_id, $data)
// Modifies the job descriptor with id $job_id, specifying new $data; returns TRUE on success, FALSE on errors.
{
	global $_AJ_error_num, $AJ_job_table_name;

	if (($old_data = AJDescRead($db_desc, $job_id)) === FALSE)
	{
		$_AJ_error_num = AJ_JOB_ID_ERROR;
		return FALSE;
	}
	$data = array_merge($old_data, $data);				// Overwrite old values with new ones
	ODBCQuery($db_desc, "UPDATE $AJ_job_table_name SET " .
			"Description = '" . addslashes($data["Description"]) . "', " .
			"ProgPath = '"  . addslashes($data["ProgPath"])  . "', " .
			"Ts = " . time() . ", " .
			"UserId = " . $data["UserId"] . " " .
			"WHERE Id = $job_id");
	if ((ODBCResult() != ODBC_NO_ERROR) || (ODBCAffectedRows($db_desc) == 0))
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return TRUE;
}



function AJDescRead(&$db_desc, $job_id)
// Returns an associative array, with the descriptor of job with id $job_id, on success, FALSE on errors.
{
	global $AJ_job_table_name, $_AJ_error_num;

	$job = ODBCSelectOneRow($db_desc, "SELECT Description, ProgPath, Ts, UserId " .
			"FROM $AJ_job_table_name WHERE Id = $job_id", ODBC_ASS);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	if (ODBCAffectedRows($db_desc) == 0)
	{
		$_AJ_error_num = AJ_JOB_ID_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $job;
}



function AJDescReadAll(&$db_desc)
// Returns an array of associative arrays, with descriptor af all jobs indexed by job id's, on success, FALSE
// on errors. Does not affect error codes.
{
	global $AJ_job_table_name, $_AJ_error_num;

	$jobs = array();
	$job = ODBCSelect($db_desc, "SELECT Id, Description, ProgPath, Ts, UserId " .
				"FROM $AJ_job_table_name", ODBC_FIRST, ODBC_ASS);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	while (ODBCAffectedRows($db_desc) > 0)
	{
		$tmp = $job["Id"];
		unset($job["Id"]);
		$jobs[$tmp] = $job;
		$job = ODBCSelect($db_desc, "", ODBC_NEXT, ODBC_ASS);
		if (ODBCResult() != ODBC_NO_ERROR)
		{
			$_AJ_error_num = AJ_DB_ERROR;
			return FALSE;
		}
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $jobs;
}



function AJDescRemove(&$db_desc, $job_id)
// Removes if possible the job descriptor with id $job_id; returns TRUE on success, FALSE on errors.
{
	global $AJ_job_table_name, $_AJ_error_num;

	ODBCQuery($db_desc, "DELETE FROM $AJ_job_table_name WHERE Id = $job_id");
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	if (ODBCAffectedRows($db_desc) == 0)
	{
		$_AJ_error_num = AJ_JOB_ID_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return TRUE;
}



function AJEnqueue(&$db_desc, $data)
// Enqueues a job, specifying its $data; returns the job queue-id on success, FALSE on errors.
{
	global $_AJ_error_num, $_AJ_def_data, $AJ_queue_table_name;

	if (($tmp = AJDescRead($db_desc, @$data["IdJob"])) === FALSE)
	{
		$_AJ_error_num = AJ_JOB_ID_ERROR;
		return FALSE;
	}
	$data = array_merge($_AJ_def_data, $data);			// Overwrite default values with new ones
	ODBCQuery($db_desc, "INSERT INTO $AJ_queue_table_name " .
			"(Hidden, IdJob, Param, Ts, UserId, Status) VALUES (" .
			($data["Hidden"] ? 1 : 0) . ", " .
			$data["IdJob"] . ", " .
			"'" . addslashes($data["Param"]) . "', " .
			time() . ", " .
			$data["UserId"] . ", " .
			AJ_READY . ")"
			);
	if ((ODBCResult() != ODBC_NO_ERROR) || (ODBCAffectedRows($db_desc) == 0))
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	list ($id) = ODBCSelectOneRow($db_desc, "SELECT LAST_INSERT_ID()");
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $id;
}



function AJEnqueuePeriodic(&$db_desc)
// Enqueues all enabled periodic job; Returns an array of job queue-id's, indexed by job periodic-id's, on
// success, FALSE on errors.
{
	global $AJ_per_table_name, $_AJ_error_num, $AJ_queue_table_name;

	$now = time();
	$p_jobs = ODBCSelectManyRows($db_desc,
			"SELECT Id AS PerId, LastQueuedTs, Period, Hidden, IdJob, Param, UserId " .
			"FROM $AJ_per_table_name " .
			"WHERE Disabled = 0 AND (LastQueuedTs + Period) < $now", ODBC_ASS);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	$result = array();
	foreach ($p_jobs as $p_job)
	{
		ODBCQuery($db_desc, "INSERT INTO $AJ_queue_table_name " .
				"(Hidden, IdJob, Param, Ts, UserId, Status) VALUES (" .
				($p_job["Hidden"] ? 1 : 0) . ", " .
				$p_job["IdJob"] . ", " .
				"'" . addslashes(_AJSubstMetaparam($db_desc, $p_job["Param"], $p_job)) . "', " .
				"$now, " .
				$p_job["UserId"] . ", " .
				AJ_READY . ")"
				);
		if ((ODBCResult() != ODBC_NO_ERROR) || (ODBCAffectedRows($db_desc) == 0))
		{
			$_AJ_error_num = AJ_DB_ERROR;
			return FALSE;
		}
		list ($id) = ODBCSelectOneRow($db_desc, "SELECT LAST_INSERT_ID()");
		if (ODBCResult() != ODBC_NO_ERROR)
		{
			$_AJ_error_num = AJ_DB_ERROR;
			return FALSE;
		}
		$result[$p_job["PerId"]] = $id;
		ODBCQuery($db_desc, "UPDATE $AJ_per_table_name SET LastQueuedTs = $now WHERE Id = $p_job[PerId]");
		if (ODBCResult() != ODBC_NO_ERROR)
		{
			$_AJ_error_num = AJ_DB_ERROR;
			return FALSE;
		}
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $result;
}



function AJErrorMsg()
// Returns the error message of the last module function called.
{
	global $_AJ_error_msg, $_AJ_error_num;

	return $_AJ_error_msg[$_AJ_error_num];
}



function AJPeriodicInsert(&$db_desc, $data)
// Inserts a periodic job, specifying its $data (the job descriptor must already exists); returns the job
// periodic-id on success, FALSE on errors.
{
	global $_AJ_error_num, $_AJ_def_data, $AJ_per_table_name;

	if (($tmp = AJDescRead($db_desc, @$data["IdJob"])) === FALSE)
	{
		$_AJ_error_num = AJ_JOB_ID_ERROR;
		return FALSE;
	}
	$data = array_merge($_AJ_def_data, array("UserId" => $tmp["UserId"]), $data);	// Overwrite default
	ODBCQuery($db_desc, "INSERT INTO $AJ_per_table_name " .							// values with new ones
			"(Period, Disabled, Hidden, IdJob, Param, Ts, UserId) VALUES (" .
			$data["Period"] . ", " .
			($data["Disabled"] ? 1 : 0) . ", " .
			($data["Hidden"] ? 1 : 0) . ", " .
			$data["IdJob"] . ", " .
			"'" . addslashes($data["Param"]) . "', " .
			time() . ", " .
			$data["UserId"] . ")"
			);
	if ((ODBCResult() != ODBC_NO_ERROR) || (ODBCAffectedRows($db_desc) == 0))
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	list ($id) = ODBCSelectOneRow($db_desc, "SELECT LAST_INSERT_ID()");
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $id;
}



function AJPeriodicModify(&$db_desc, $per_id, $data)
// Modifies the periodic job with periodic-id $per_id, specifying new $data (the job descriptor must already
// exists); returns TRUE on success, FALSE
// on errors.
{
	global $_AJ_error_num, $AJ_per_table_name;

	if (($old_data = AJPeriodicRead($db_desc, $per_id)) === FALSE)
	{
		$_AJ_error_num = AJ_PER_ID_ERROR;
		return FALSE;
	}
	$data = array_merge($old_data, $data);				// Overwrite old values with new ones
	if (($tmp = AJDescRead($db_desc, $data["IdJob"])) === FALSE)
	{
		$_AJ_error_num = AJ_JOB_ID_ERROR;
		return FALSE;
	}
	ODBCQuery($db_desc, "UPDATE $AJ_per_table_name SET " .
			"Period = " . $data["Period"] . ", " .
			"Disabled = " . ($data["Disabled"] ? 1 : 0) . ", " .
			"Hidden = " . ($data["Hidden"] ? 1 : 0) . ", " .
			"IdJob = " . $data["IdJob"] . ", " .
			"Param = '" . addslashes($data["Param"]) . "', " .
			"Ts = " . time() . ", " .
			"UserId = " . $data["UserId"] . " " .
			"WHERE Id = $per_id"
			);
	if ((ODBCResult() != ODBC_NO_ERROR) || (ODBCAffectedRows($db_desc) == 0))
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return TRUE;
}



function AJPeriodicRead(&$db_desc, $per_id)
// Returns an associative array with data of the periodic job with periodic-id $per_id on success, FALSE on
// errors. Does not affect error codes.
{
	global $AJ_per_table_name, $_AJ_error_num;

	$p_job = ODBCSelectOneRow($db_desc,
				"SELECT LastQueuedTs, Period, Disabled, Hidden, IdJob, Param, Ts, UserId " .
				"FROM $AJ_per_table_name WHERE Id = $per_id", ODBC_ASS);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	if (ODBCAffectedRows($db_desc) == 0)
	{
		$_AJ_error_num = AJ_PER_ID_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $p_job;
}



function AJPeriodicReadAll(&$db_desc)
// Returns an array of associative arrays with data af all periodic jobs, indexed by job periodic-id's on
// success, FALSE on errors. Does not affect error codes.
{
	global $AJ_per_table_name, $_AJ_error_num;

	$p_jobs = array();
	$p_job = ODBCSelect($db_desc,
				"SELECT Id, LastQueuedTs, Period, Disabled, Hidden, IdJob, Param, Ts, UserId " .
				"FROM $AJ_per_table_name", ODBC_FIRST, ODBC_ASS);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	while (ODBCAffectedRows($db_desc) > 0)
	{
		$tmp = $p_job["Id"];
		unset($p_job["Id"]);
		$p_jobs[$tmp] = $p_job;
		$p_job = ODBCSelect($db_desc, "", ODBC_NEXT, ODBC_ASS);
		if (ODBCResult() != ODBC_NO_ERROR)
		{
			$_AJ_error_num = AJ_DB_ERROR;
			return FALSE;
		}
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $p_jobs;
}



function AJPeriodicRemove(&$db_desc, $per_id)
// Removes the periodic job with periodic-id $per_id; returns TRUE on success, FALSE on errors.
{
	global $AJ_per_table_name, $_AJ_error_num;

	ODBCQuery($db_desc, "DELETE FROM $AJ_per_table_name WHERE Id = $per_id");
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	if (ODBCAffectedRows($db_desc) == 0)
	{
		$_AJ_error_num = AJ_PER_ID_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return TRUE;
}



function AJQueuedFlushEnded(&$db_desc, $ok_timeout = AJ_OK_TIMEOUT, $err_timeout = AJ_ERR_TIMEOUT)
// Removes ended jobs still in execution queue, if they are ended for a time greater than the specified
// timeout (timeouts can be different for jobs normally ended and jobs ended with errors). Returns an
// array of removed jobs data, indexed by job queue-id's, on success, FALSE on errors.
{
	global $AJ_queue_table_name, $_AJ_error_num;

	$now = time();
	$q_jobs = ODBCSelectManyRows($db_desc,
			"SELECT Id, Hidden, IdJob, Param, Ts, UserId, Status, Result, Output, TsStart, Elapsed " .
			"FROM $AJ_queue_table_name WHERE " .
			"Status = " . AJ_ENDED . " AND (" .
				"(Result = 0 AND (TsStart + Elapsed + $ok_timeout) < $now) OR " .
				"(Result <> 0 AND (TsStart + Elapsed + $err_timeout) < $now))", ODBC_ASS);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	$result = array();
	foreach ($q_jobs as $q_job)
	{
		$tmp = $q_job["Id"];
		unset($q_job["Id"]);
		$result[$tmp] = $q_job;
		ODBCQuery($db_desc, "DELETE FROM $AJ_queue_table_name WHERE Id = $tmp");
		if (ODBCResult() != ODBC_NO_ERROR)
		{
			$_AJ_error_num = AJ_DB_ERROR;
			return FALSE;
		}
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $result;
}



function AJQueuedRead(&$db_desc, $queue_id)
// Returns an associative array with data of the queued job with queue-id $queue_id on success, FALSE on
// errors.
{
	global $AJ_queue_table_name, $_AJ_error_num;

	$q_job = ODBCSelectOneRow($db_desc,
				"SELECT Hidden, IdJob, Param, Ts, UserId, Status, Result, Output, TsStart, Elapsed " .
				"FROM $AJ_queue_table_name WHERE Id = $queue_id", ODBC_ASS);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	if (ODBCAffectedRows($db_desc) == 0)
	{
		$_AJ_error_num = AJ_QUEUE_ID_ERROR;
		return FALSE;
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $q_job;
}



function AJQueuedReadAll(&$db_desc)
// Returns an array of associative arrays with data af all queued jobs, indexed by job queue-id's on
// success, FALSE on errors.
{
	global $AJ_queue_table_name, $_AJ_error_num;

	$q_jobs = array();
	$q_job = ODBCSelect($db_desc,
				"SELECT Id, Hidden, IdJob, Param, Ts, UserId, Status, Result, Output, TsStart, Elapsed " .
				"FROM $AJ_queue_table_name", ODBC_FIRST, ODBC_ASS);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	while (ODBCAffectedRows($db_desc) > 0)
	{
		$tmp = $q_job["Id"];
		unset($q_job["Id"]);
		$q_jobs[$tmp] = $q_job;
		$q_job = ODBCSelect($db_desc, "", ODBC_NEXT, ODBC_ASS);
		if (ODBCResult() != ODBC_NO_ERROR)
		{
			$_AJ_error_num = AJ_DB_ERROR;
			return FALSE;
		}
	}
	$_AJ_error_num = AJ_NO_ERROR;
	return $q_jobs;
}



function AJQueuedRun(&$db_desc)
// Runs queued ready jobs ("Status" == AJ_READY), passing them all to AJ_SCHEDULED and executing one at
// the time ("Status" == AJ_RUNNING) until ended ("Status" == AJ_ENDED). Returns an array of executed jobs
// data, indexed by job queue-id's, on success, FALSE on errors.
{
	global $AJ_cfg_table_name, $AJ_queue_table_name, $_AJ_error_num, $AJ_job_table_name, $AJ_db_log,
			$AJ_log_table_name;

	$cfg = array("CfgTableName"	=> $AJ_cfg_table_name);
	OCInit($cfg, $db_desc);
	if (! OCUTsl($cfg, "AJQueuedRun", 5))
	{
		$_AJ_error_num = AJ_ALREADY_RUNNING;
		return FALSE;
	}
	ODBCQuery($db_desc, "UPDATE $AJ_queue_table_name SET Status = " . AJ_SCHEDULED . " WHERE Status = " .
			AJ_READY);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	$scheduled = ODBCSelectManyRows($db_desc,
			"SELECT q.Id AS QueueId, q.Hidden, q.IdJob AS JobId, j.ProgPath, q.Param, q.UserId " .
			"FROM $AJ_queue_table_name AS q LEFT JOIN $AJ_job_table_name AS j ON q.IdJob = j.Id " .
			"WHERE q.Status = " . AJ_SCHEDULED, ODBC_ASS);
	if (ODBCResult() != ODBC_NO_ERROR)
	{
		$_AJ_error_num = AJ_DB_ERROR;
		return FALSE;
	}
	$return_data = array();
	foreach ($scheduled as $s_job)
	{
		$param = _AJSubstMetaparam($db_desc, $s_job["Param"], $s_job);
		$cmd_line = "$s_job[ProgPath] " . $param;
		$output = array();
		$ts_start = time();
		ODBCQuery($db_desc, "UPDATE $AJ_queue_table_name SET " .
				"Status = " . AJ_RUNNING . ", " .
				"TsStart = $ts_start " .
				"WHERE Id = " . $s_job["QueueId"]);
		if (ODBCResult() != ODBC_NO_ERROR)
		{
			$_AJ_error_num = AJ_DB_ERROR;
			return FALSE;
		}
		DUTimerReset("AJQueuedRun");
		exec($cmd_line, $output, $result);
		$elapsed = DUTimerRead("AJQueuedRun");
		$output = implode("\n", $output);
		ODBCQuery($db_desc, "UPDATE $AJ_queue_table_name SET " .
				"Status = " . AJ_ENDED . ", " .
				"Result = " . $result . ", " .
				"Output = '" . addslashes($output) . "', " .
				"Elapsed = " . $elapsed . " " .
				"WHERE Id = " . $s_job["QueueId"]);
		if (ODBCResult() != ODBC_NO_ERROR)
		{
			$_AJ_error_num = AJ_DB_ERROR;
			return FALSE;
		}
		$return_data[$s_job["QueueId"]] = array("Hidden" => $s_job["Hidden"], "JobId" => $s_job["JobId"],
				"ProgPath" => $s_job["ProgPath"], "Param" => $s_job["Param"], "UserId" => $s_job["UserId"],
				"Result" => $result, "Output" => $output, "TsStart" => $ts_start, "Elapsed" => $elapsed);
		if ($AJ_db_log)
			DLAppend($AJ_log_table_name, $s_job["Hidden"], $s_job["ProgPath"], $param, $s_job["UserId"],
					$result, $output, $ts_start, $elapsed);
	}
	OCUUnlock($cfg, "AJQueuedRun");
	OCDone($cfg);
	$_AJ_error_num = AJ_NO_ERROR;
	return $return_data;
}



function AJResult()
// Returns the error code of the last module function called.
{
	global $_AJ_error_num;

	return $_AJ_error_num;
}



//--- Private functions code -----------------------------------------------------------------------



function _AJSubstMetaparam(&$db_desc, $param, $desc = array())
// Substitutes all meta parameters in $param, as specified in Configuration table ($AJ_cfg_table_name)
// section "MetaParam" ($AJ_meta_sec_name), with available values in the job or period descriptor $desc.
{
	global $AJ_cfg_table_name, $AJ_meta_sec_name;

	$cfg = array("CfgTableName"	=> $AJ_cfg_table_name);
	OCInit($cfg, $db_desc);
	$metas = OCReadSection($cfg, $AJ_meta_sec_name, OC_ASS);
	foreach ($metas as $key => $meta)
		if (array_key_exists($key, $desc))
			$param = str_replace($meta, $desc[$key], $param);
	OCDone($cfg);
	return $param;
}



//--- Initialization code --------------------------------------------------------------------------



?>
